from shutil import which

from ..repo.github import GitHubRepo
from ..utils import run_cmd
from .base import Script


class GitHubScript(Script[GitHubRepo]):
    """Subclass this to define how to select and modify GitHub repositories.
    Uses the GitHub CLI, which must already be installed and logged in.

    :param submit: Whether to submit the changes. This is disabled by default,
        to give you a chance to develop the changes first.
    :param orgs: The list of users/orgs to clone repositories from.
    """

    orgs: list[str]
    """The list of GitHub users/orgs to clone repositories from."""

    def __init__(self, *, submit: bool = False, orgs: list[str] | None = None) -> None:
        super().__init__(submit=submit)

        if orgs is not None:
            self.orgs = orgs

    def list_all_repos(self) -> list[GitHubRepo]:
        return [
            GitHubRepo(self, org, name)
            for org in self.orgs
            for name in run_cmd(
                which("gh"),  # type: ignore[arg-type]
                "repo",
                "list",
                "--no-archived",
                "--json",
                "name",
                "--jq",
                ".[] | .name",
                org,
            ).stdout.splitlines()
        ]



from __future__ import annotations

import typing as t
from functools import cached_property
from pathlib import Path
from shutil import which
from subprocess import CompletedProcess

from ..repo.git import GitRepo
from ..utils import run_cmd

if t.TYPE_CHECKING:
    from ..script.base import Script


class GitHubRepo(GitRepo):
    """Subclass this to define how to manipulate Git repositories. This extends
    the plain :class:`GitRepo` to clone and create PRs using the GitHub CLI.

    :param script: The script being run to modify this repo.
    :param org: The GitHub user/org that owns the repo.
    :param name: The GitHub repo name.
    """

    _gh_exe: str | None = which("gh")
    direct_submit: bool = False
    """Whether to merge and push directly to the target branch, rather than
    creating a PR. This is disabled by default as a PR will give more
    opportunity to review any mistakes with the automated changes.
    """

    def __init__(self, script: Script[t.Any], org: str, name: str) -> None:
        self.org = org
        self.name = name
        super().__init__(script=script, remote_id=self.full_name)

    def gh_cmd(self, *args: str | Path) -> CompletedProcess[str]:
        """Call and pass args to the `gh` command.

        :param args: Command line arguments to the `gh` command.
        """
        if self._gh_exe is None:
            raise RuntimeError("GitHub CLI is not installed.")

        return run_cmd(self._gh_exe, *args)

    @cached_property
    def full_name(self) -> str:
        """The `org/name` identifier for the repo."""
        return f"{self.org}/{self.name}"

    def clone(self) -> None:
        self.local_dir.parent.mkdir(parents=True, exist_ok=True)
        self.gh_cmd(
            "repo",
            "clone",
            self.full_name,
            self.local_dir,
            "--",
            "-b",
            self.script.target,
        )

    def submit(self) -> None:
        if self.direct_submit:
            super().submit()
            return

        self.git_cmd("push", "--set-upstream", "origin", self.script.branch)
        self.gh_cmd(
            "pr",
            "create",
            "--base",
            self.script.target,
            "--title",
            self.script.title,
            "--body",
            self.script.body,
        )



import inspect
import typing as t
from functools import cached_property
from os import PathLike
from pathlib import Path

import jinja2
import platformdirs

from ..repo.base import Repo
from ..utils import read_text
from ..utils import wrap_text


class Script[RepoType: Repo]:
    """Defines how to select repositories and modify them. Typically, you'll
    want to subclass a more specific class that is already set up to work with a
    remote host. You'd subclass this class if you wanted to define such a class
    for a new host.

    :param submit: Whether to submit the changes. This is disabled by default,
        to give you a chance to develop the changes first.
    """

    target: str = "main"
    """The name of the target branch to branch off of and merge into."""

    branch: str
    """The name of the work branch to create."""

    title: str
    """A short title describing the change. Used as the first line of the
    automatic commit, as well as the title of the PR. By convention, this should
    be at most 50 characters.
    """

    body: str
    """Additional description about the change. Used in the commit message
    after the title, separated by an empty line. Also used as the body of the
    PR. This will be re-wrapped to 72 characters to match convention.
    """

    def __init__(self, *, submit: bool = False) -> None:
        if self.target == self.branch:
            raise ValueError(
                "Work branch name must be different than target branch name."
            )

        source_file = inspect.getsourcefile(self.__class__)

        if source_file is None:
            raise RuntimeError("Could not determine script root.")

        self.root_dir: Path = Path(source_file).parent
        """The directory containing the running script. Used to reference
        resource files and templates.
        """

        self.jinja_env: jinja2.Environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(self.root_dir),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )
        """A Jinja environment configured to use :attr:`root_dir` as a template
        folder. See :meth:`render_template`.
        """

        self.clones_dir: Path = platformdirs.user_cache_path("modify-repos") / "clones"
        """Directory where repos are cloned to. Uses the appropriate user cache
        dir for the platform.
        """

        self.clones_dir.mkdir(parents=True, exist_ok=True)

        if not (ignore := self.clones_dir / ".gitignore").exists():
            ignore.write_text("*\n")

        self.body = wrap_text(self.body, width=72)
        self.enable_submit = submit
        """Whether to submit the changes. This is disabled by default, to give
        you a chance to develop the changes first. It is set from the `submit`
        param.
        """

    def render_template(self, name: str, /, **kwargs: t.Any) -> str:
        """Render the named template file with context. Uses :attr:`jinja_env`,
        which finds templates next to the script file.

        :param name: Template name to load.
        :param kwargs: Context to pass to the render call.
        """
        return self.jinja_env.get_template(name).render(**kwargs)

    def list_all_repos(self) -> list[RepoType]:
        """Get the list of all repos that may be cloned. Override this to
        define how to generate this list. Called by :meth:`list_repos`."""
        raise NotImplementedError

    def list_repos(self) -> list[RepoType]:
        """Get the filtered list of repos that will be cloned. Override
        :meth:`list_all_repos` and :meth:`select_for_clone` to control what is
        returned here. Called by :meth:`run`."""
        return [r for r in self.list_all_repos() if self.select_for_clone(r)]

    @cached_property
    def full_target(self) -> str:
        """The upstream target branch, which is :attr:`target` prefixed by
        `origin/`.
        """
        return f"origin/{self.target}"

    @cached_property
    def commit_message(self) -> str:
        """The message to use for the automatic commit in :meth:`commit`.
        Defaults to :attr:`title` and :attr:`body` separated by a blank line.
        """
        return f"{self.title}\n\n{self.body}"

    def select_for_clone(self, repo: Repo) -> bool:
        """Select what repos are returned by :meth:`list_repos`. Each repo from
        :meth:`list_all_repos` is passed, and will be used if this method returns
        true for it.

        For example, override this to return true if the repo name matches a set
        of names.

        :param repo: The repo to filter.
        """
        return True

    def select_for_modify(self, repo: Repo) -> bool:
        """Select whether :meth:`modify` will be called on the repo. Called by
        :meth:`run` while the current directory is the cloned repo dir.

        For example, override this to return false if the repo does not contain
        a file to be removed, or already contains a file to be added.

        :param repo: The repo to filter.
        """
        return True

    def modify(self, repo: Repo) -> None:
        """Perform modifications to the repo. Called by :meth:`run` while the
        current directory is the cloned repo dir.

        If this leaves uncommitted changes, :meth:`.Repo.commit_if_needed` will
        detect that and commit automatically. You can also add and commit
        manually to skip that behavior.

        :param repo: The repo to modify.
        """
        raise NotImplementedError

    def run(self) -> None:
        """Call :meth:`.Repo.run` for each selected repo."""
        for repo in self.list_repos():
            repo.run()

    def read_text(self, path: str | PathLike[str], strip: bool = True) -> str:
        """Read a text file, where `path` is relative to the script's directory
        :attr:`root_dir`. The file will be read as UTF-8.

        :param path: Path to file to read. Relative paths are relative to
            :attr:`root_dir`.
        :param strip: Strip leading and trailing empty spaces and lines. Enabled
            by default. The text is often formatted into an existing file, so
            stripping spaces makes working with it more predictable.
        """
        return read_text(self.root_dir / path, strip=strip)



from __future__ import annotations

import os
import shlex
import subprocess
import textwrap
import typing as t
from inspect import cleandoc
from os import PathLike
from pathlib import Path

import click


def run_cmd(
    *args: str | PathLike[str], **kwargs: t.Any
) -> subprocess.CompletedProcess[str]:
    """Wrapper around :meth:`subprocess.run`. Args are passed positionally
    rather than as a list. Stdout and stderr are combined, and use text mode.
    The initial command is echoed, and if the return code is not 0, the output
    and code are echoed.

    :param args: Command line arguments.
    :param kwargs: Other arguments passed to `subprocess.run`.
    """
    s_args = [os.fspath(v) if isinstance(v, PathLike) else v for v in args]
    click.echo(f"$ {shlex.join(s_args)}")
    result: subprocess.CompletedProcess[str] = subprocess.run(
        s_args,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        **kwargs,
    )

    if result.returncode:
        click.echo(result.stdout)
        click.secho(f"exited with code {result.returncode}", fg="red")

    return result


def wrap_text(text: str, width: int = 80) -> str:
    """Wrap a multi-line, multi-paragraph string. The text is dedented and empty
    spaces and lines are stripped, to support triple-quoted strings. Paragraphs
    are separated by a blank line `\\\\n\\\\n`. Tabs are converted to 4 spaces,
    and very long words are not wrapped.

    :param text: The text to process.
    :param width: The number of characters to wrap at.
    """
    return "\n\n".join(
        textwrap.fill(p, width=width, tabsize=4, break_long_words=False)
        for p in cleandoc(text).split("\n\n")
    )


def read_text(path: str | PathLike[str], strip: bool = True) -> str:
    """Read a file as UTF-8 text.

    :param path: The path to the file to read.
    :param strip: Whether to strip empty spaces and lines. Enabled by default.
        This makes the text more predictable to work with when inserting it into
        another text.
    """
    text = Path(path).read_text("utf8")

    if strip:
        text = text.strip()

    return text


def write_text(path: str | PathLike[str], text: str, end_nl: bool = True) -> None:
    """Write a file as UTF-8 text.

    :param path: The path to the file to write.
    :param text: The text to write.
    :param end_nl: Whether to ensure the text ends with exactly one newline
        `\\\\n`. This keeps file endings consistent.
    """
    if end_nl:
        text = f"{text.rstrip('\n')}\n"

    Path(path).write_text(text, "utf8")



from __future__ import annotations

from pathlib import Path
from shutil import which
from subprocess import CompletedProcess

from ..utils import run_cmd
from .base import Repo


class GitRepo(Repo):
    """Subclass this to define how to manipulate Git repositories.

    :param script: The script being run to modify this repo.
    """

    _git_exe = which("git")
    add_untracked: bool = False
    """Whether to consider untracked files when checking if there are changes
    and adding files in the auto :meth:`commit`. By default this is false to
    avoid accidentally adding generated files, but this means you need to
    remember to call :meth:`git_add` for any completely new files.
    """

    def git_cmd(self, *args: str | Path) -> CompletedProcess[str]:
        """Call and pass args to the `git` command.

        :param args: Command line arguments to the `git` command.
        """
        if self._git_exe is None:
            raise RuntimeError("Git is not installed.")

        return run_cmd(self._git_exe, *args)

    def reset_target(self) -> None:
        self.git_cmd("switch", "-f", self.script.target)
        self.git_cmd("reset", "--hard", self.script.full_target)
        self.git_cmd("pull", "--prune")

    def reset_branch(self) -> None:
        self.git_cmd("switch", "-C", self.script.branch, self.script.target)

    def needs_commit(self) -> bool:
        args = ["status", "--porcelain"]

        if not self.add_untracked:
            args.append("--untracked-files=no")

        return bool(self.git_cmd(*args).stdout)

    def auto_commit(self) -> None:
        if self.add_untracked:
            self.add_files(all=True)
        else:
            self.add_files(update=True)

        self.git_cmd("commit", "--message", self.script.commit_message)

    def needs_submit(self) -> bool:
        return bool(self.git_cmd("cherry", self.script.full_target).stdout)

    def submit(self) -> None:
        self.git_cmd("switch", self.script.target)
        self.git_cmd("merge", "--ff-only", self.script.branch)
        self.git_cmd("push")

    def add_files(
        self, *items: str | Path, update: bool = False, all: bool = False
    ) -> None:
        """Call `git add`.

        :param items: Files to add or update. Can be empty.
        :param update: Update all tracked files.
        :param all: Add all files, including untracked, excluding ignored.
        """
        if all:
            self.git_cmd("add", "--all")

        if update:
            self.git_cmd("add", "--update")

        self.git_cmd("add", *items)

    def rm_files(self, *items: str | Path) -> None:
        """Call `git rm` for any given files that exist. Missing files are
        skipped so that the command doesn't return an error.

        :param items: Files to delete if they exist.
        """
        to_remove = [item for item in items if Path(item).exists()]

        if to_remove:
            self.git_cmd("rm", *to_remove)



from .repo.base import Repo
from .repo.github import GitHubRepo
from .script.base import Script
from .script.github import GitHubScript

__all__ = [
    "GitHubRepo",
    "GitHubScript",
    "Repo",
    "Script",
]



from __future__ import annotations

import typing as t
from contextlib import chdir
from functools import cached_property
from pathlib import Path

import click

if t.TYPE_CHECKING:
    from ..script.base import Script


class Repo:
    """Defines how to manipulate a repository. :meth:`.Script.list_repos` will
    return instances of a subclass of this that are set up to work with a
    specific type of repo, and most of the methods will be called as part of
    running the script. You'd subclass this class if you wanted to define such a
    class for a new type of repo.

    :param script: The script being run to modify this repo.
    :param remote_id: A value that identifies where this repo was cloned from.
    """

    remote_id: str

    def __init__(self, script: Script[t.Any], remote_id: str) -> None:
        self.script = script
        """The script being run to modify this repo."""

        self.remote_id = remote_id
        """A value that identifies where this repo was cloned from. The format
        depends on how the script finds repos.
        """

    @cached_property
    def local_dir(self) -> Path:
        """The path where this repo is cloned."""
        return self.script.clones_dir / self.remote_id

    def clone(self) -> None:
        """Clone the repository unconditionally. This is called by
        :meth:`clone_if_needed`.
        """
        raise NotImplementedError

    def clone_if_needed(self) -> None:
        """Clone the repository if the local directory doesn't exist. Calls
        :meth:`clone`. Called by :meth:`run`.
        """
        if not self.local_dir.exists():
            self.clone()

    def reset_target(self) -> None:
        """Reset the base branch that will be branched off of and merged into.
        This should ensure the repository is clean and up to date, discarding
        any changes from previous unsuccessful runs. Called by :meth:`run`.
        """
        raise NotImplementedError

    def reset_branch(self) -> None:
        """Create or reset the work branch. This should ensure the branch is
        freshly created from the target branch. Called by :meth:`run`.
        """
        raise NotImplementedError

    def needs_commit(self) -> bool:
        """Check if there are uncommitted changes. Called by
        :meth:`auto_commit_if_needed`.
        """
        raise NotImplementedError

    def auto_commit(self) -> None:
        """Create a commit unconditionally. This is called by
        :meth:`auto_commit_if_needed` to create the automatic commit when there
        are uncommitted changes, and should add those changes to the commit. It
        should not be called to create other intermediate commits. It should use
        :attr:`.Script.message` as the commit message.
        """
        raise NotImplementedError

    def auto_commit_if_needed(self) -> None:
        """Create a commit if there are uncommitted changes. Calls
        :meth:`needs_commit` and :meth:`auto_commit_if_needed`. Called by
        :meth:`run`.
        """
        if self.needs_commit():
            self.auto_commit()

    def needs_submit(self) -> bool:
        """Check if there are commits that have not been pushed upstream. Called
        by :meth:`submit_if_needed`."""
        raise NotImplementedError

    def submit(self) -> None:
        """Submit the changes upstream. What this means depends on the
        implementation; whether it merges, creates a PR, or something else.
        """
        raise NotImplementedError

    def submit_if_needed(self) -> None:
        """Submit the changes if there are any changes. Is disabled by default
        by :attr:`.Script.enable_submit`, to prevent accidental submission of a
        script in development. Calls :meth:`needs_submit` and :meth:`submit`.
        Called by :meth:`run`.
        """
        if self.script.enable_submit and self.needs_submit():
            self.submit()
        else:
            click.secho("skipping submit", fg="yellow")

    def run(self) -> None:
        """Run the full workflow for this repo: clone, reset, modify, commit,
        submit. Calls many of the other methods defined in this class. Called
        by :meth:`.Script.run`.
        """
        click.secho(self.remote_id, fg="green")
        self.clone_if_needed()

        with chdir(self.local_dir):
            self.reset_target()

            if not self.script.select_for_modify(self):
                click.secho("skipping modify", fg="yellow")
                return

            self.reset_branch()
            self.script.modify(self)
            self.auto_commit_if_needed()
            self.submit_if_needed()
